VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "PortMapper"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit
'
'================
'PortMapper v 1.6
'================
'
'A class that wraps the Windows UPnP Library for mapping ports through
'UPnP-enabled NAT routers.
'
'The router must support UPnP and NAT port mapping (sometimes referred
'to as "Virtual Server" mapping).
'
'The OS must be Windows XP or later.  It must have the UPnP Framework
'installed (normally a standard feature).  This is sometimes shown in
'Windows as "Network Discovery" or as "SSPD" and "UPnP" (two items).
'
'The software firewall (such as Windows Firewall) must allow UPnP to
'pass through and the NAT router must have UPnP enabled.
'
'
'PROPERTIES
'
'   ExternalIP As String R/O
'
'       After calling OpenMapper() and once the Opened event has been
'       raised, this property contains the public IP address of the
'       NAT router.
'
'   InternalIP As String R/O
'
'       After calling OpenMapper() this property contains the private
'       IP address that was passed in via the OpenMapper() call.
'
'METHODS
'
'   OpenMapper(ByVal LocalIP As String)
'
'       Before the other methods can be called you must first call
'       OpenMapper() passing in your private IP address.
'
'       You can obtain your private IP address from a Winsock control's
'       LocalIP property.
'
'   CloseMapper()
'
'       Stops any active AsyncFind operation and releases UPnPLib
'       resources.  After calling CloseMapper() you can call
'       OpenMapper() again if necessary.
'
'   Add(ByVal Description As String,
'       ByVal Protocol As MSWinsockLib.ProtocolConstants,
'       ByVal ExternalPort As Long,
'       Optional ByVal InternalPort As Long = -1,
'       Optional ByVal LeaseDuration As Long = 0
'       Optional ByVal Enabled As Boolean = True)
'
'       Attempts to add, renew, or modify a port mapping.
'
'       Description
'           A string meant to describe this mapping, e.g. "HTTP" or
'           "Dark Moon Chat" and so on.
'
'       Protocol
'           TCP or UDP, values the same as used by the Winsock
'           control: 0 = sckTCPProtocol and 1 = sckUDPProtocol.
'
'       ExternalPort
'           Port number to be presented at the NAT router's public
'           interface, i.e. the port clients will connect to.
'
'       InternalPort
'           Port number that the server behind the NAT router will
'           listen for TCP connections or receive UDP datagrams on.
'           This can be the same value as ExternalPort.  if allowed
'           to default PortMapper will set it = to ExternalPort.
'
'           Some routers REQUIRE that the port numbers be the same!
'
'       LeaseDuration
'           The amount of time (in seconds) that the NAT router
'           should maintain this port mapping.  After expiration the
'           NAT router can remove the mapping automatically.
'
'           A value of 0 implies a static, i.e. permanent mapping.
'           Not all NAT routers support dynamic mapping, and will
'           therefore require a 0 for this value.
'
'           When using dynamic mappings a program can call Add()
'           again to renew the mapping for another interval.  This
'           must be done a few seconds prior to expiration to avoid
'           interruptions in service.  All parameters must be
'           supplied again to do this renewal.
'
'       Enabled
'           Specify whether to establish the port mapping enabled or
'           disabled.  If you create a disabled port mapping you must
'           call Add() again with all parameters to enable it.
'
'       Some erroneous conditions will result in an exception being
'       thrown.  Sometimes a NAT router can get into a state where it
'       won't accept port mapping actions and they will fail until
'       after the router is rebooted.  Most NAT routers wil only
'       support a finite number of mappings.
'
'       Mappings established via UPnP are not shown in the "remote
'       console" web pages of all NAT routers.  Most routers will make
'       log entries for UPnP port mapping events though.
'
'       In cases where the ExternalPort and Protocol are the same as an
'       existing mapping but InternalIP is different, Add() results in
'       an exception.  You cannot map the same external Port & Protocol
'       to two different internal machines.
'
'   Remove(ByVal ExternalPort As Long,
'          ByVal Protocol As MSWinsockLib.ProtocolConstants)
'
'       Attempts to remove an existing port mapping.
'
'       ExternalPort
'           The port number presented at the NAT router's public
'           interface for this mapping.
'
'       Protocol
'           The protocol for this mapping.
'
'       Some erroneous conditions will result in an exception being
'       thrown.  Sometimes a NAT router can get into a state where it
'       won't accept port mapping actions and they will fail until
'       after the router is rebooted.
'
'       If the mapping doesn't exist an exception is raised.
'
'EVENTS
'
'   Opened(ByVal Success As Boolean)
'
'       This event is raised after OpenMapper() has been called and
'       the NAT router's WANIPConnection has been located.  After this
'       event the other methods may be called and the ExternalIP and
'       InternalIP properties are valid.
'
'       If Success is False when this event is raised it means that
'       the WANIPConnection Service could not be located, so the
'       ExternalIP property won't be correct and you cannot call
'       Add() or Remove().
'
'       Some resources are tied up until after CloseMapper() has been
'       called.
'
'       The UPnP "search" process takes some time, perhaps even as long
'       as 10 to 30 seconds!  In most cases if it takes more than 10
'       seconds it means:
'
'           o There is no UPnP capable NAT router in the network.  This
'             can also mean the local system is directly attached to
'             the public Internet.
'           o The NAT router has UPnP disabled.
'           o A software firewall is blocking UPnP.
'
'       To detect this the program can use a Timer control with
'       Interval set to 10000 (10 seconds).  Set Enabled = True just
'       before calling OpenMapper(), and in the Opened event set
'       Enabled = False.
'
'       If the Timer's Timer event occurs you know you have one of the
'       conditions above.  Set Enabled = False, call CloseMapper, and
'       warn the user or proceed assuming you are directly connected
'       to the Internet.
'

Private Const CONN_SVCTYPEID_URI As String = "urn:schemas-upnp-org:service:WANIPConnection:1"
Private Const CONN_ID_URI As String = "urn:upnp-org:serviceId:WANIPConn1"
Private Const vbUInteger As Integer = 18
Private Const vbULong As Integer = 19

Private Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" ( _
    ByVal Destination As Long, ByVal Source As Long, _
    ByVal Length As Long)

Private UDFinder As UPNPLib.UPnPDeviceFinder
Private WithEvents PMCBs As PortMapperCBs
Attribute PMCBs.VB_VarHelpID = -1
Private CONNService As UPNPLib.UPnPService
Private findData As Long
Private mExternalIP As String
Private mInternalIP As String
Private blnSuccess As Boolean
Private blnOpen As Boolean
Private intHackUI2 As Integer
Private intHackUI4 As Integer

Public Event Opened(ByVal Success As Boolean)

Public Property Get ExternalIP() As String
    ExternalIP = mExternalIP
End Property

Public Property Get InternalIP() As String
    InternalIP = mInternalIP
End Property

Public Sub Add( _
    ByVal Description As String, _
    ByVal Protocol As MSWinsockLib.ProtocolConstants, _
    ByVal ExternalPort As Long, _
    Optional ByVal InternalPort As Long = -1, _
    Optional ByVal LeaseDuration As Long = 0, _
    Optional ByVal Enabled As Boolean = True)

    Dim varInActionArgs, varOutActionArgs

    If InternalPort < 0 Then InternalPort = ExternalPort
    If blnOpen Then
        ReDim varInActionArgs(0 To 7)
        varInActionArgs(0) = ""
        varInActionArgs(1) = CLng(ExternalPort)
        varInActionArgs(2) = IIf(Protocol = sckTCPProtocol, "TCP", "UDP")
        varInActionArgs(3) = CLng(InternalPort)
        varInActionArgs(4) = mInternalIP
        varInActionArgs(5) = Enabled
        varInActionArgs(6) = Description
        varInActionArgs(7) = LeaseDuration
        CopyMemory VarPtr(varInActionArgs(1)), VarPtr(intHackUI2), 2
        CopyMemory VarPtr(varInActionArgs(3)), VarPtr(intHackUI2), 2
        CopyMemory VarPtr(varInActionArgs(7)), VarPtr(intHackUI4), 2
        ReDim varOutActionArgs(0 To 0)
        CONNService.InvokeAction "AddPortMapping", _
                                 varInActionArgs, _
                                 varOutActionArgs
    Else
        Err.Raise &H80049B00, "PortMapper", "PortMapper must be open before calling Add"
    End If
End Sub

Public Sub CloseMapper()
    If Not PMCBs Is Nothing Then Set PMCBs = Nothing
    If Not UDFinder Is Nothing Then
        On Error Resume Next
        UDFinder.CancelAsyncFind findData
        On Error GoTo 0
        Set UDFinder = Nothing
    End If
    If Not CONNService Is Nothing Then Set CONNService = Nothing
    blnOpen = False
End Sub

Public Sub OpenMapper(ByVal LocalIP As String)
    blnSuccess = False
    mInternalIP = LocalIP
    mExternalIP = ""
    intHackUI2 = vbUInteger
    intHackUI4 = vbULong
    Set UDFinder = New UPNPLib.UPnPDeviceFinder
    Set PMCBs = New PortMapperCBs
    findData = CallByName(UDFinder, "CreateAsyncFind", VbMethod, CONN_SVCTYPEID_URI, 0, PMCBs)
    UDFinder.StartAsyncFind findData
End Sub

Public Sub Remove(ByVal ExternalPort As Long, ByVal Protocol As MSWinsockLib.ProtocolConstants)
    Dim varInActionArgs, varOutActionArgs
    
    If blnOpen Then
        ReDim varInActionArgs(0 To 2)
        varInActionArgs(0) = ""
        varInActionArgs(1) = CLng(ExternalPort)
        varInActionArgs(2) = IIf(Protocol = sckTCPProtocol, "TCP", "UDP")
        CopyMemory VarPtr(varInActionArgs(1)), VarPtr(intHackUI2), 2
        ReDim varOutActionArgs(0 To 0)
        CONNService.InvokeAction "DeletePortMapping", _
                                 varInActionArgs, _
                                 varOutActionArgs
    Else
        Err.Raise &H80049B02, "PortMapper", "PortMapper must be open before calling Remove"
    End If
End Sub

Private Sub Class_Terminate()
    CloseMapper
End Sub

Private Sub PMCBs_DeviceAdded(ByVal Device As UPNPLib.IUPnPDevice)
    Dim Services As UPNPLib.UPnPServices
    Dim Service As UPNPLib.UPnPService
    Dim varInActionArgs, varOutActionArgs
    
    On Error Resume Next
    Set Services = Device.Services
    If Err.Number = 0 Then
        On Error GoTo 0
        With Services
            If .Count > 0 Then
                On Error Resume Next
                Set Service = .Item(CONN_ID_URI)
                If Err.Number = 0 Then
                    On Error GoTo 0
                    Set CONNService = Service
                    ReDim varInActionArgs(0 To 0)
                    ReDim varOutActionArgs(0 To 0)
                    CONNService.InvokeAction "GetExternalIPAddress", _
                                             varInActionArgs, _
                                             varOutActionArgs
                    mExternalIP = varOutActionArgs(0)
                    blnSuccess = True
                    blnOpen = True
                Else
                    On Error GoTo 0
                End If
            End If
        End With
    Else
        On Error GoTo 0
    End If
    
    UDFinder.CancelAsyncFind findData
    RaiseEvent Opened(blnSuccess)
End Sub

Private Sub PMCBs_SearchComplete()
    If Not blnSuccess Then
        RaiseEvent Opened(False)
    End If
End Sub
